@@ -9,6 +9,7 @@ module Control.Monad.Par.Scheds.DirectInternal where
 import Control.Applicative
 import "mtl" Control.Monad.Cont as C
 import qualified "mtl" Control.Monad.Reader as RD
+import Control.Monad.IO.Class (liftIO)
 
 import qualified System.Random.MWC as Random
 
@@ -18,16 +19,22 @@ import Data.IORef
 import qualified Data.Set as S
 import Data.Word (Word64)
 import Data.Concurrent.Deque.Class (WSDeque)
+import Control.Monad.Fix (MonadFix (mfix))
+#if MIN_VERSION_base(4,4,0)
+import GHC.IO.Unsafe (unsafeDupableInterleaveIO)
+#else
+import GHC.IO.Unsafe (unsafeInterleaveIO)
+#endif
 
 #ifdef USE_CHASELEV
 #warning "Note: using Chase-Lev lockfree workstealing deques..."
 import Data.Concurrent.Deque.ChaseLev.DequeInstance
 import Data.Concurrent.Deque.ChaseLev as R
-#else
-import Data.Concurrent.Deque.Reference.DequeInstance
-import Data.Concurrent.Deque.Reference as R
 #endif
 
+import Control.Exception (Exception, throwIO, BlockedIndefinitelyOnMVar (..),
+                          catch)
+
 -- Our monad stack looks like this:
 --      ---------
 --        ContT
@@ -44,6 +51,33 @@ newtype Par a = Par { unPar :: C.ContT () ROnly a }
     deriving (Functor, Applicative, Monad, MonadCont, RD.MonadReader Sched)
 type ROnly = RD.ReaderT Sched IO
 
+instance MonadFix Par where
+  mfix = fixPar
+
+-- | Take the monadic fixpoint of a 'Par' computation. This is
+-- the definition of 'mfix' for 'Par'. Throws 'FixParException'
+-- if the result is demanded strictly within the computation.
+fixPar :: (a -> Par a) -> Par a
+-- We do this IO-style, rather than ST-style, in order to get a
+-- consistent exception type. Using the ST-style mfix, a strict
+-- argument could lead us to *either* a <<loop>> exception *or*
+-- (if the wrong sort of computation gets re-run) a "multiple-put"
+-- error.
+fixPar f = Par $ ContT $ \ar -> RD.ReaderT $ \sched -> do
+  mv <- newEmptyMVar
+  ans <- unsafeDupableInterleaveIO (readMVar mv `catch`
+      \ ~BlockedIndefinitelyOnMVar -> throwIO FixParException)
+  flip RD.runReaderT sched $
+    runContT (unPar (f ans)) $ \a -> liftIO (putMVar mv a) >> ar a
+
+#if !MIN_VERSION_base(4,4,0)
+unsafeDupableInterleaveIO :: IO a -> IO a
+unsafeDupableInterleaveIO = unsafeInterleaveIO
+#endif
+
+data FixParException = FixParException deriving Show
+instance Exception FixParException
+
 type SessionID = Word64
 
 -- An ID along with a flag to signal completion:
@@ -112,10 +146,12 @@ modifyHotVar_ v fn = atomicModifyIORef v (\a -> (fn a, ()))
 readHotVar    = readIORef
 writeHotVar   = writeIORef
 instance Show (IORef a) where
-  show ref = "<ioref>"
+  show _ref = "<ioref>"
 
+writeHotVarRaw :: HotVar a -> a -> IO ()
 -- hotVarTransaction = id
 hotVarTransaction = error "Transactions not currently possible for IO refs"
+readHotVarRaw :: HotVar a -> IO a
 readHotVarRaw  = readHotVar
 writeHotVarRaw = writeHotVar
 
@@ -130,7 +166,7 @@ modifyHotVar_ v fn = modifyMVar_ v (return . fn)
 readHotVar    = readMVar
 writeHotVar v x = do swapMVar v x; return ()
 instance Show (MVar a) where
-  show ref = "<mvar>"
+  show _ref = "<mvar>"
 
 -- hotVarTransaction = id
 -- We could in theory do this by taking the mvar to grab the lock.
