@@ -14,7 +14,7 @@ module Control.Monad.Par.Scheds.TraceInternal (
    runPar, runParIO, runParAsync,
    -- runParAsyncHelper,
    new, newFull, newFull_, get, put_, put,
-   pollIVar, yield,
+   pollIVar, yield, fixPar, FixParException (..)
  ) where
 
 
@@ -22,9 +22,17 @@ import Control.Monad as M hiding (mapM, sequence, join)
 import Prelude hiding (mapM, sequence, head,tail)
 import Data.IORef
 import System.IO.Unsafe
+#if MIN_VERSION_base(4,4,0)
+import GHC.IO.Unsafe (unsafeDupableInterleaveIO)
+#else
+import GHC.IO.Unsafe (unsafeInterleaveIO)
+#endif
 import Control.Concurrent hiding (yield)
 import GHC.Conc (numCapabilities)
 import Control.DeepSeq
+import Control.Monad.Fix (MonadFix (mfix))
+import Control.Exception (Exception, throwIO, BlockedIndefinitelyOnMVar (..),
+                          catch)
 -- import Text.Printf
 
 #if !MIN_VERSION_base(4,8,0)
@@ -97,6 +105,9 @@ sched _doSync queue t = loop t
         r <- io
         loop (c r)
 
+data FixParException = FixParException deriving Show
+instance Exception FixParException
+
 -- | Process the next item on the work queue or, failing that, go into
 --   work-stealing mode.
 reschedule :: Sched -> IO ()
@@ -182,6 +193,31 @@ instance Applicative Par where
    (<*>) = ap
    pure a = Par ($ a)
 
+instance MonadFix Par where
+   mfix = fixPar
+
+-- | Take the monadic fixpoint of a 'Par' computation. This is
+-- the definition of 'mfix' for 'Par'. Throws 'FixParException'
+-- if the result is demanded strictly within the computation.
+fixPar :: (a -> Par a) -> Par a
+-- We do this IO-style, rather than ST-style, in order to get a
+-- consistent exception type. Using the ST-style mfix, a strict
+-- argument could lead us to *either* a <<loop>> exception *or*
+-- (if the wrong sort of computation gets re-run) a "multiple-put"
+-- error.
+fixPar f = Par $ \ c ->
+  LiftIO (do
+    mv <- newEmptyMVar
+    ans <- unsafeDupableInterleaveIO (readMVar mv
+             `catch` \ ~BlockedIndefinitelyOnMVar -> throwIO FixParException)
+    case f ans of
+      Par q -> pure $ q $ \a -> LiftIO (putMVar mv a) (\ ~() -> c a)) id
+
+#if !MIN_VERSION_base(4,4,0)
+unsafeDupableInterleaveIO :: IO a -> IO a
+unsafeDupableInterleaveIO = unsafeInterleaveIO
+#endif
+
 newtype IVar a = IVar (IORef (IVarContents a))
 -- data IVar a = IVar (IORef (IVarContents a))
 
@@ -189,9 +225,8 @@ newtype IVar a = IVar (IORef (IVarContents a))
 instance Eq (IVar a) where
   (IVar r1) == (IVar r2) = r1 == r2
 
--- Forcing evaluation of a IVar is fruitless.
 instance NFData (IVar a) where
-  rnf _ = ()
+  rnf !_ = ()
 
 
 -- From outside the Par computation we can peek.  But this is nondeterministic.
@@ -281,7 +316,14 @@ new  = Par $ New Empty
 
 -- | Creates a new @IVar@ that contains a value
 newFull :: NFData a => a -> Par (IVar a)
-newFull x = deepseq x (Par $ New (Full x))
+-- What are we doing here? We're manually raising the arity
+-- of newFull from 2 to 3, which seems like it's probably what
+-- we want most of the time. Notably, fmapping over the result
+-- gives really awful-looking Core if we don't do this.
+-- Regardless, I think we logically want to force the
+-- value when it's installed in the IVar rather than
+-- when we create the action to install it in the IVar.
+newFull x = Par $ \c -> x `deepseq` New (Full x) c
 
 -- | Creates a new @IVar@ that contains a value (head-strict only)
 newFull_ :: a -> Par (IVar a)
@@ -309,7 +351,11 @@ put_ v !a = Par $ \c -> Put v a (c ())
 -- Sometimes partial strictness is more appropriate: see 'put_'.
 --
 put :: NFData a => IVar a -> a -> Par ()
-put v a = deepseq a (Par $ \c -> Put v a (c ()))
+-- Manually raise the arity, which seems likely to be what
+-- we want most of the time. We really want to force the
+-- value when it's installed in the IVar, not when we
+-- create the Par action to install it in the IVar.
+put v a = Par $ \c -> a `deepseq` Put v a (c ())
 
 -- | Allows other parallel computations to progress.  (should not be
 -- necessary in most cases).
